﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  07/01/13    Created 
//
//---------------------------------------------------------------------------

namespace System.Data.Entity
{
    using System.Collections;
    using System.Collections.Generic;
    using System.ComponentModel.DataAnnotations.Schema;
    using System.Linq;
    using System.Linq.Expressions;

    public static class DatabaseExtensions
    {
        /// <summary>
        /// Creates a unique non-clustered database index.
        /// </summary>
        /// <typeparam name="TModel">The type of model to create the index for.</typeparam>
        /// <param name="database">The database to create an index in.</param>
        /// <param name="expression">An lambda expression selecting the fields to create an index for.</param>
        public static void CreateUniqueNonClusteredIndex<TModel>(this Database database, params Expression<Func<TModel, object>>[] expression)
        {
            CreateIndex<TModel>(database, "UNIQUE NONCLUSTERED", expression);
        }

        /// <summary>
        /// Creates a clustered database index.
        /// </summary>
        /// <typeparam name="TModel">The type of model to create the index for.</typeparam>
        /// <param name="database">The database to create an index in.</param>
        /// <param name="expression">An lambda expression selecting the fields to create an index for.</param>
        public static void CreateClusteredIndex<TModel>(this Database database, params Expression<Func<TModel, object>>[] expression)
        {
            CreateIndex<TModel>(database, "CLUSTERED", expression);
        }

        /// <summary>
        /// Creates a unique, clustered database index.
        /// </summary>
        /// <typeparam name="TModel">The type of model to create the index for.</typeparam>
        /// <param name="database">The database to create an index in.</param>
        /// <param name="expression">An lambda expression selecting the fields to create an index for.</param>
        public static void CreateUniqueClusteredIndex<TModel>(this Database database, params Expression<Func<TModel, object>>[] expression)
        {
            CreateIndex<TModel>(database, "UNIQUE CLUSTERED", expression);
        }

        /// <summary>
        /// Creates an index.
        /// </summary>
        /// <typeparam name="TModel">The type of model to create the index for.</typeparam>
        /// <param name="database">The database to create an index in.</param>
        /// <param name="verb">The type of index to create.</param>
        /// <param name="expression">An lambda expression selecting the fields to create an index for.</param>
        private static void CreateIndex<TModel>(this Database database, string verb, params Expression<Func<TModel, object>>[] expression)
        {
            if (database == null)
                throw new ArgumentNullException("database");

            if (verb == null)
                throw new ArgumentNullException("verb");

            //get lambda expression names:
            string[] names = expression.Select(e => LambdaExpressionName(e.Body)).ToArray();

            if (names.Length > 0)
            {
                Type type = typeof(TModel);
                //get any table name:
                Attribute[] attribute = Attribute.GetCustomAttributes(type).Where(t => t is TableAttribute).ToArray();

                string name;
                if (attribute.Length == 0) //if no explicit table name, use the type
                    name = type.Name;
                else
                    name = ((TableAttribute)attribute[0]).Name; //use the table name

                //execute the sql command:
                database.ExecuteSqlCommand("CREATE " + verb + " INDEX IX_" + name + "__" + string.Join("_", names) + " ON " + name + " (" + string.Join(",", names) + ")");
            }
        }

        /// <summary>
        /// Attempts to find the lambda expression member name.
        /// </summary>
        /// <param name="expression">The lambda expression.</param>
        /// <returns>The lambda expression member name.</returns>
        private static string LambdaExpressionName(Expression expression)
        {
            //convert to a MemberExpression
            MemberExpression memberExpression = expression as MemberExpression;

            //if conversion failed
            if (memberExpression == null)
            {
                //convert to UnaryExpression
                UnaryExpression unaryExpression = expression as UnaryExpression;

                //if conversion successful
                if (unaryExpression != null)
                    memberExpression = unaryExpression.Operand as MemberExpression; //set the memberExpression to the UnaryExpression operand
            }

            if (memberExpression == null)
                throw new ArgumentException("Could not find name from expression.", "expression");

            return memberExpression.Member.Name; //return the expressions member name
        }
    }
}
