﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  07/01/13    Created 
//  12/30/14    Added support for extended format
//---------------------------------------------------------------------------

namespace System.Data.Entity
{
    using System.ComponentModel.DataAnnotations.Schema;
    using System.Configuration;
    using System.Data.Entity.ModelConfiguration;
    using System.Data.SqlClient;
    using System.IO;
    using System.Linq;
    using System.Net;
    using System.Net.Sockets;
    using System.Web;
    using System.Web.Caching;

    public class IPCountryCodeContext : DbContext
    {
        #region Constants
        /// <summary>
        /// Defines the connection key for the IPCountryCodeContext database.
        /// </summary>
        private const string CONNECTION_KEY = "IPCountryCodeConnection";
        #endregion

        #region IPv6RegionalInternetRegistryDataEntityTypeConfiguration Class
        /// <summary>
        /// The IPv6RegionalInternetRegistryDataEntityTypeConfiguration configuration class configures the IPv6RegionalInternetRegistryData model.
        /// </summary>
        private class IPv6RegionalInternetRegistryDataEntityTypeConfiguration : EntityTypeConfiguration<IPv6RegionalInternetRegistryData>
        {
            public IPv6RegionalInternetRegistryDataEntityTypeConfiguration()
            {
                this.HasKey(k => new { k.StartIP1, k.StartIP2, k.EndIP1, k.EndIP2 });
            }
        }
        #endregion

        #region Entity Collections
        /// <summary>
        /// Gets or sets a collection of regional internet registries.
        /// </summary>
        public DbSet<RegionalInternetRegistry> RegionalInternetRegistries { get; set; }
        /// <summary>
        /// Gets or sets the IPv4 regional registry data.
        /// </summary>
        public DbSet<IPv4RegionalInternetRegistryData> IPv4 { get; set; }
        /// <summary>
        /// Gets or sets the IPv6 regional registry data.
        /// </summary>
        public DbSet<IPv6RegionalInternetRegistryData> IPv6 { get; set; }
        #endregion

        #region Constructor
        /// <summary>
        /// Initializes the IPCountryCodeContext DbContext class.
        /// </summary>
        protected IPCountryCodeContext()
            : base(CONNECTION_KEY)
        {

        }
        #endregion

        #region Overrides
        /// <summary>
        /// Creates the model.
        /// </summary>
        /// <param name="modelBuilder">The builder that defines the model for the context being created.</param>
        protected override void OnModelCreating(DbModelBuilder modelBuilder)
        {
            //removes the StartIP column identity:
            modelBuilder.Entity<IPv4RegionalInternetRegistryData>().HasKey(p => p.StartIP).Property(p => p.StartIP).HasDatabaseGeneratedOption(DatabaseGeneratedOption.None);
            //adds IPv6 entity configuration to create a key for start and end columns
            modelBuilder.Configurations.Add(new IPv6RegionalInternetRegistryDataEntityTypeConfiguration());

            base.OnModelCreating(modelBuilder);
        }
        #endregion

        #region Static Methods
        /// <summary>
        /// Gets the connection string to the IPCountryCodeContext database.
        /// </summary>
        protected static string GetConnectionString()
        {
            ConnectionStringSettings setting = ConfigurationManager.ConnectionStrings[CONNECTION_KEY];

            if (setting == null)
                throw new SettingsPropertyNotFoundException("The IPCountryCodeConnection connection string was not found or is invalid.");

            return setting.ConnectionString;
        }

        #region Updating Registry Data
        /// <summary>
        /// Downloads and updates all regional internet registry data.
        /// </summary>
        public static void UpdateRegionalInternetRegistryData()
        {
            RegionalInternetRegistry[] registries;

            using (IPCountryCodeContext context = new IPCountryCodeContext())
                registries = context.RegionalInternetRegistries.ToArray();

            foreach (RegionalInternetRegistry registry in registries)
                UpdateRegionalInternetRegistryFile(registry);
        }

        /// <summary>
        /// Downloads and loads the specified regional internet registry file.
        /// </summary>
        /// <param name="registry"></param>
        private static void UpdateRegionalInternetRegistryFile(RegionalInternetRegistry registry)
        {
            string tempPath = null;

            try
            {
                tempPath = Path.GetTempFileName(); //create temp file

                try
                {
                    using (WebClient client = new WebClient())
                        client.DownloadFile(registry.URL, tempPath); //download file to temp
                }
                catch
                {
                    //log exception
                    return;
                }

                LoadRegionalInternetRegistryFile(registry, tempPath);
            }
            finally
            {
                if (tempPath != null && File.Exists(tempPath))
                    File.Delete(tempPath);
            }
        }

        /// <summary>
        /// Parses and loads a regional internet registry file into the database.
        /// </summary>
        /// <param name="registry">The regional internet registry to update in the database.</param>
        /// <param name="path">The path to a regional internet registry file.</param>
        private static void LoadRegionalInternetRegistryFile(RegionalInternetRegistry registry, string path)
        {
            using (DataTable ipv6Table = new DataTable("IPv4RIRData"))
            using (DataColumn ipv6CountryParam = new DataColumn("CountryCode", typeof(string)))
            using (DataTable ipv4Table = new DataTable("IPv4RIRData"))
            using (DataColumn ipv4CountryParam = new DataColumn("CountryCode", typeof(string)))
            {
                //add ipv6 table columns
                ipv6Table.Columns.Add("StartIP1", typeof(long));
                ipv6Table.Columns.Add("StartIP2", typeof(long));
                ipv6Table.Columns.Add("EndIP1", typeof(long));
                ipv6Table.Columns.Add("EndIP2", typeof(long));

                ipv6CountryParam.MaxLength = 2;
                ipv6Table.Columns.Add(ipv6CountryParam);

                //add ipv4 table columns
                ipv4Table.Columns.Add("StartIP", typeof(int));
                ipv4Table.Columns.Add("EndIP", typeof(int));

                ipv4CountryParam.MaxLength = 2;
                ipv4Table.Columns.Add(ipv4CountryParam);

                //open the RIR file
                using (StreamReader reader = new StreamReader(path))
                {
                    for (; ; ) //continuous
                    {
                        //read each line
                        string line = reader.ReadLine();

                        if (line == null) //EOF
                            break;

                        //registry|cc|type|start|value|date|status
                        string[] split = line.Split('|');

                        int l = split.Length;

                        if (l != 7 && l != 8) //allow extended format
                            continue;

                        switch (split[6]) //status
                        {
                            case "allocated": //only take allocated data, the assigned data may contain duplicates

                                //number of ipv4 addresses or number of ipv6 bits
                                int num_end;
                                if (!int.TryParse(split[4], out num_end)) //value
                                    continue;

                                switch (split[2].ToLower()) //type
                                {
                                    case "ipv4": //accept IPv4 and IPv6 entries
                                    case "ipv6":

                                        IPAddress address;
                                        //attempt to parse the ip address:
                                        if (!IPAddress.TryParse(split[3], out address)) //start
                                            continue;

                                        //get ip as binary:
                                        byte[] start = address.GetAddressBytes();

                                        switch (address.AddressFamily)
                                        {
                                            case AddressFamily.InterNetwork:
                                                num_end--;

                                                if (BitConverter.IsLittleEndian)
                                                    Array.Reverse(start);

                                                //convert into a signed value:
                                                int startip = (int)(BitConverter.ToUInt32(start, 0) - int.MaxValue);

                                                //add entry to table
                                                ipv4Table.Rows.Add(startip, startip + num_end, split[1]);
                                                break;
                                            case AddressFamily.InterNetworkV6:
                                                byte[] end = new byte[16];
                                                Array.Copy(start, end, 16);

                                                byte sub;
                                                int index;

                                                for (int i = 127; i >= num_end; i--) //set each bit
                                                {
                                                    sub = (byte)(i % 8); //get which octet
                                                    index = (i - sub) / 8; //get which byte
                                                    end[index] |= (byte)(Math.Pow(2, 7 - sub)); //set flag
                                                }

                                                if (BitConverter.IsLittleEndian)
                                                {
                                                    Array.Reverse(start, 0, 8);
                                                    Array.Reverse(start, 8, 8);
                                                    Array.Reverse(end, 0, 8);
                                                    Array.Reverse(end, 8, 8);
                                                }

                                                //add entry to table
                                                ipv6Table.Rows.Add((long)(BitConverter.ToUInt64(start, 0) - long.MaxValue), (long)(BitConverter.ToUInt64(start, 8) - long.MaxValue), (long)(BitConverter.ToUInt64(end, 0) - long.MaxValue), (long)(BitConverter.ToUInt64(end, 8) - long.MaxValue), split[1]);
                                                break;
                                        }

                                        break;
                                }
                                break;
                        }
                    }
                }

                InsertIPv4RIRData(registry.RegionalInternetRegistryID, ipv4Table);
                InsertIPv6RIRData(registry.RegionalInternetRegistryID, ipv6Table);
            }
        }

        /// <summary>
        /// Inserts IPv6 regional internet registry.
        /// </summary>
        /// <param name="regionalInternetRegistryID">The regional internet registry id</param>
        /// <param name="ipv6Data">IPv6 data table to insert into the database.</param>
        private static void InsertIPv6RIRData(int regionalInternetRegistryID, DataTable ipv6Data)
        {
            using (SqlConnection conn = new SqlConnection(GetConnectionString()))
            using (SqlCommand cmd = new SqlCommand("InsertIPv6RIRData", conn))
            {
                conn.Open();
                cmd.CommandType = CommandType.StoredProcedure;
                cmd.Parameters.AddWithValue("@RIRID", regionalInternetRegistryID); //add the regional internet registry id
                cmd.Parameters.AddWithValue("@RIRTable", ipv6Data); //add the IPv6 data table

                cmd.ExecuteNonQuery();
            }
        }

        /// <summary>
        /// Inserts IPv4 regional internet registry.
        /// </summary>
        /// <param name="regionalInternetRegistryID">The regional internet registry id</param>
        /// <param name="ipv4Data">IPv4 data table to insert into the database.</param>
        private static void InsertIPv4RIRData(int regionalInternetRegistryID, DataTable ipv4Data)
        {
            using (SqlConnection conn = new SqlConnection(GetConnectionString()))
            using (SqlCommand cmd = new SqlCommand("InsertIPv4RIRData", conn))
            {
                conn.Open();
                cmd.CommandType = CommandType.StoredProcedure;
                cmd.Parameters.AddWithValue("@RIRID", regionalInternetRegistryID); //add the regional internet registry id
                cmd.Parameters.AddWithValue("@RIRTable", ipv4Data); //add the IPv4 data table

                cmd.ExecuteNonQuery();
            }
        }
        #endregion

        #region Static Lookup Methods
        /// <summary>
        /// Gets the country code for the specified address.
        /// </summary>
        /// <param name="ipString">The address to look up.</param>
        /// <returns>A two character country code or null.</returns>
        public static string LookupCountryCode(string ipString)
        {
            if (ipString == null)
                return null;

            //get cache time in milliseconds
            string cacheTimeoutValue = ConfigurationManager.AppSettings["IPCountryCodeContext_IPAddressCacheTimeout"];

            if (cacheTimeoutValue != null)
            {
                int cacheTimeout;
                if (int.TryParse(cacheTimeoutValue, out cacheTimeout) && cacheTimeout > 0) //if caching is available
                {
                    HttpContext context = HttpContext.Current; //get http context

                    if (context != null) //only cache with context
                    {
                        string cacheKey = NetCacheKey.CreateCacheKey("RegionalInternetRegistry", ipString); //create a cache key for the ip address
                        string countryCode = context.Cache[cacheKey] as string; //get any cached country code

                        if (countryCode == null)
                        {
                            countryCode = LookupCountryCodeFromDatabase(ipString); //lookup ip address

                            if (countryCode == null) //store empty string if no country code found
                                countryCode = string.Empty;

                            //add ip address to cache:
                            context.Cache.Add(cacheKey, countryCode, null, Cache.NoAbsoluteExpiration, TimeSpan.FromMilliseconds(cacheTimeout), CacheItemPriority.Low, null);
                        }

                        if (countryCode.Length == 0) //empty string from cache indicates no country code was found in the database
                            return null;

                        return countryCode;
                    }
                }
            }

            return LookupCountryCodeFromDatabase(ipString); //lookup and return ip address
        }

        /// <summary>
        /// Gets the country code for the specified address from the database.
        /// </summary>
        /// <param name="ipString">The address to look up.</param>
        /// <returns>A two character country code or null.</returns>
        private static string LookupCountryCodeFromDatabase(string ipString)
        {
            IPAddress address;
            if (!IPAddress.TryParse(ipString, out address))
                return null;

            byte[] ipAddressBytes = address.GetAddressBytes();

            switch (address.AddressFamily)
            {
                case AddressFamily.InterNetwork: //IPv4
                    {
                        if (BitConverter.IsLittleEndian)
                            Array.Reverse(ipAddressBytes);

                        //convert into a signed value:
                        int ip = (int)(BitConverter.ToUInt32(ipAddressBytes, 0) - int.MaxValue);

                        using (IPCountryCodeContext context = new IPCountryCodeContext())
                            //find the country code for an entry within the IP range
                            return (from a in context.IPv4
                                    where a.StartIP <= ip && a.EndIP >= ip
                                    select a.CountryCode).FirstOrDefault();
                    }
                case AddressFamily.InterNetworkV6: //IPv6
                    {
                        if (BitConverter.IsLittleEndian)
                        {
                            Array.Reverse(ipAddressBytes, 0, 8);
                            Array.Reverse(ipAddressBytes, 8, 8);
                        }

                        //split the 128-bit value into two 64-bit signed values:
                        long ip1 = (long)(BitConverter.ToUInt64(ipAddressBytes, 0) - long.MaxValue);
                        long ip2 = (long)(BitConverter.ToUInt64(ipAddressBytes, 8) - long.MaxValue);

                        using (IPCountryCodeContext context = new IPCountryCodeContext())
                            //find the country code for an entry within the IP range
                            return (from a in context.IPv6
                                    where (ip1 > a.StartIP1 || (ip1 == a.StartIP1 && ip2 >= a.StartIP2)) && (ip1 < a.EndIP1 || (ip1 == a.EndIP1 && ip2 <= a.EndIP2))
                                    select a.CountryCode).FirstOrDefault();
                    }
            }

            return null;
        }
        #endregion
        #endregion
    }
}
