﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  07/01/13    Created 
//  12/30/14    Added extended format URLs
//
//---------------------------------------------------------------------------

namespace System.Web.Mvc
{
    using System.Data.Entity;
    using System.Net;

    /// <summary>
    /// Defines an initializer for the <see cref="IPCountryCodeContext"/> class.
    /// </summary>
    public sealed class IPCountryCodeInitializer : CreateDatabaseIfNotExists<IPCountryCodeContext>
    {
        /// <summary>
        /// Seeds the database by setting recovery model, creating indexes, adding table types and preloading registries.
        /// </summary>
        /// <param name="context">The context to seed.</param>
        protected override void Seed(IPCountryCodeContext context)
        {
            //create a non-clustered index for the EndIP:
            context.Database.CreateUniqueNonClusteredIndex<IPv4RegionalInternetRegistryData>(x => x.EndIP);

            //use simple recovery model:
            context.Database.ExecuteSqlCommand(@"
DECLARE @alter AS NVARCHAR(256);
SET @alter = 'alter database [' + DB_NAME() + '] set recovery simple';
exec(@alter);");

            //create the table type to bulk add IPv4 data:
            context.Database.ExecuteSqlCommand(@"
CREATE TYPE [dbo].[IPv4RIRTable] AS TABLE (
  [StartIP] int NOT NULL,
  [EndIP] int NOT NULL,
  [CountryCode] char(2) COLLATE SQL_Latin1_General_CP1_CI_AS NOT NULL,
  PRIMARY KEY CLUSTERED ([StartIP], [EndIP]),
  UNIQUE ([EndIP]),
  UNIQUE ([StartIP])
)");

            //create the table type to bulk add IPv6 data:
            context.Database.ExecuteSqlCommand(@"
CREATE TYPE [dbo].[IPv6RIRTable] AS TABLE (
  [StartIP1] bigint NOT NULL,
  [StartIP2] bigint NOT NULL,
  [EndIP1] bigint NOT NULL,
  [EndIP2] bigint NOT NULL,
  [CountryCode] char(2) COLLATE SQL_Latin1_General_CP1_CI_AS NOT NULL
)");

            //create a stored procedure to add IPv4 data table:
            context.Database.ExecuteSqlCommand(@"
CREATE PROCEDURE dbo.InsertIPv4RIRData
@RIRID int,
@RIRTable IPv4RIRTable readonly
AS
SET XACT_ABORT ON;
BEGIN
	BEGIN TRANSACTION;
	DELETE FROM IPv4 WHERE RegionalInternetRegistry_RegionalInternetRegistryID=@RIRID;
	INSERT INTO IPv4 (RegionalInternetRegistry_RegionalInternetRegistryID, StartIP, EndIP, CountryCode) SELECT @RIRID, StartIP, EndIP, CountryCode FROM @RIRTable;
	COMMIT TRANSACTION;
END");

            //create a stored procedure to add IPv6 data table:
            context.Database.ExecuteSqlCommand(@"
CREATE PROCEDURE dbo.InsertIPv6RIRData
@RIRID int,
@RIRTable IPv6RIRTable readonly
AS
SET XACT_ABORT ON;
BEGIN
	BEGIN TRANSACTION;
	DELETE FROM IPv6 WHERE RegionalInternetRegistry_RegionalInternetRegistryID=@RIRID;
	INSERT INTO IPv6 (RegionalInternetRegistry_RegionalInternetRegistryID, StartIP1, StartIP2, EndIP1, EndIP2, CountryCode) SELECT @RIRID, StartIP1, StartIP2, EndIP1, EndIP2, CountryCode FROM @RIRTable;
	COMMIT TRANSACTION;
END");

            //preload regional internet registries:
            context.RegionalInternetRegistries.Add(new RegionalInternetRegistry { URL = "ftp://ftp.ripe.net/ripe/stats/delegated-ripencc-latest" });
            context.RegionalInternetRegistries.Add(new RegionalInternetRegistry { URL = "ftp://ftp.lacnic.net/pub/stats/lacnic/delegated-lacnic-latest" });
            context.RegionalInternetRegistries.Add(new RegionalInternetRegistry { URL = "ftp://ftp.arin.net/pub/stats/arin/delegated-arin-extended-latest" });
            context.RegionalInternetRegistries.Add(new RegionalInternetRegistry { URL = "ftp://ftp.apnic.net/pub/apnic/stats/apnic/delegated-apnic-extended-latest" });
            context.RegionalInternetRegistries.Add(new RegionalInternetRegistry { URL = "ftp://ftp.afrinic.net/pub/stats/afrinic/delegated-afrinic-latest" });

            base.Seed(context);
        }
    }
}